"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DingzAccessory = void 0;
const simple_color_converter_1 = __importDefault(require("simple-color-converter"));
const qs_1 = __importDefault(require("qs"));
const semver_1 = __importDefault(require("semver"));
const limit_number_1 = __importDefault(require("limit-number"));
// Internal types
const settings_1 = require("./settings");
const dingzTypes_1 = require("./lib/dingzTypes");
const commonTypes_1 = require("./lib/commonTypes");
const errors_1 = require("./lib/errors");
const dingzDaBaseAccessory_1 = require("./lib/dingzDaBaseAccessory");
/**
 * Platform Accessory
 * An instance of this class is created for each accessory your platform registers
 * Each accessory may expose multiple services of different service types.
 */
class DingzAccessory extends dingzDaBaseAccessory_1.DingzDaBaseAccessory {
    /**
     * DingzAccessory constructor
     * @param _platform platform pointer
     * @param _accessory accessory pointer
     */
    constructor(_platform, _accessory) {
        super(_platform, _accessory);
        this._platform = _platform;
        this._accessory = _accessory;
        this.mustInit = false;
        // Todo: Make proper internal representation
        this.dingzStates = {
            // Outputs
            Dimmers: [],
            WindowCovers: [],
            LED: {
                on: false,
                hsv: '0;0;100',
                rgb: 'FFFFFF',
                mode: 'hsv',
            },
            // Inputs
            Buttons: {
                '1': { event: commonTypes_1.ButtonAction.SINGLE_PRESS, state: dingzTypes_1.ButtonState.OFF },
                '2': { event: commonTypes_1.ButtonAction.SINGLE_PRESS, state: dingzTypes_1.ButtonState.OFF },
                '3': { event: commonTypes_1.ButtonAction.SINGLE_PRESS, state: dingzTypes_1.ButtonState.OFF },
                '4': { event: commonTypes_1.ButtonAction.SINGLE_PRESS, state: dingzTypes_1.ButtonState.OFF },
            },
            // Sensors
            Temperature: 0,
            Motion: false,
            Brightness: 0,
        };
        this.configTimestamp = 0;
        // _accessory will contain the current config
        // regardless of wheter it is a new or a restored accessory
        this.config = this.accessory.context.config;
        this.hw = this.device.hwInfo;
        if (this.config && this.hw) {
            // only run 'reconfigureAccessory()' if we have a config
            this.log.info('Config available, initialization started');
            this.mustInit = false;
            this.reconfigureAccessory(true);
        }
        else {
            this.log.warn('Config not available, initialization deferred');
            this.mustInit = true;
        }
        // Remove Reachability service if still present
        const bridgingService = this.accessory.getService(this.platform.Service.BridgingState);
        if (bridgingService) {
            this.accessory.removeService(bridgingService);
        }
    }
    setOutputHandlers(outputService, index) {
        outputService
            .getCharacteristic(this.platform.Characteristic.On)
            .on("set" /* CharacteristicEventTypes.SET */, this.setOn.bind(this, index)) // SET - bind to the `setOn` method below
            .on("get" /* CharacteristicEventTypes.GET */, this.getOn.bind(this, index));
        this.eb.on("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */, this.updateDimmerState.bind(this, outputService, index));
    }
    /**
     * Set the accessory information
     * - Firmware
     * - Manufacturer
     * - S/N etc.
     */
    reconfigureAccessory(init = false) {
        var _a, _b;
        // Set base info
        // Persist updated info
        // Sanity check for "empty" SerialNumber
        this.log.debug(`Attempting to set SerialNumber (which can not be empty) -> front_sn: <${this.hw.front_sn}>`);
        const serialNumber = !this.hw.front_sn || '' === this.hw.front_sn
            ? this.device.mac
            : this.hw.front_sn; // MAC will always be defined for a correct device
        this.log.debug(`Setting SerialNumber (which can not be empty) -> : <${serialNumber}>`);
        this.accessory
            .getService(this.platform.Service.AccessoryInformation)
            .setCharacteristic(this.platform.Characteristic.Manufacturer, 'iolo AG')
            .setCharacteristic(this.platform.Characteristic.AppMatchingIdentifier, 'ch.iolo.dingz.consumer')
            // Update info from deviceInfo
            .setCharacteristic(this.platform.Characteristic.ConfiguredName, this.accessory.displayName)
            .setCharacteristic(this.platform.Characteristic.Name, this.accessory.displayName)
            .setCharacteristic(this.platform.Characteristic.Model, this.device.model)
            .setCharacteristic(this.platform.Characteristic.FirmwareRevision, (_a = this.hw.fw_version) !== null && _a !== void 0 ? _a : 'Unknown')
            .setCharacteristic(this.platform.Characteristic.HardwareRevision, (_b = this.hw.hw_version_puck) !== null && _b !== void 0 ? _b : 'Unknown')
            .setCharacteristic(this.platform.Characteristic.SerialNumber, serialNumber);
        // Add Dimmers, Blinds etc.
        this.log.info('Adding output devices for ', this.device.address, ' -> [...]');
        this.setButtonCallbacks();
        // Now we have what we need and can create the services …
        this.configureOutputs(init);
        this.configureBlinds(init);
        this.configureButtons(init);
        // Add auxiliary services (Motion, Temperature)
        if (this.hw.has_pir) {
            // dingz has a Motion sensor -- let's create it
            this.addMotionService();
        }
        else {
            this.log.info('This dingz has no Motion sensor.');
            this.removeMotionService();
        }
        // dingz has a temperature sensor and an LED,
        // make these available here
        this.addTemperatureService();
        this.addLEDService();
        this.addLightSensorService();
    }
    setButtonCallbacks() {
        // Only necessary for firmware version < 1.2.x
        if (semver_1.default.lt(this.hw.fw_version, '1.2.0')) {
            this.log.debug('Enable PIR callback for older firmware revisions');
            this.enablePIRCallback();
        }
        this.getButtonCallbackUrl()
            .then((callBackUrl) => {
            var _a, _b;
            // Set the callback URL
            const endpoints = ['generic'];
            const platformCallbackUrl = this.platform.getCallbackUrl();
            // Add PIR callbacks, depending on dingz Firmware version
            if (this.hw.has_pir) {
                if (semver_1.default.lt(this.hw.fw_version, '1.2.0')) {
                    endpoints.push('pir/single');
                }
                else if (semver_1.default.lt(this.hw.fw_version, '1.4.0')) {
                    endpoints.push('pir/generic', 'pir/rise', 'pir/fall');
                }
                else {
                    // FIXES #511: Newer FW have (yet!) other endpoint for PIR callbacks
                    endpoints.push('pir1/rise', 'pir1/fall');
                }
            }
            if (this.platform.config.callbackOverride) {
                this.log.warn('Override callback URL ->', callBackUrl);
                this.platform.setButtonCallbackUrl({
                    baseUrl: this.baseUrl,
                    token: this.device.token,
                    endpoints: endpoints,
                });
            }
            else if (
            // FIXME: because of #511
            (semver_1.default.lt(this.hw.fw_version, '1.4.0') &&
                !((_a = callBackUrl === null || callBackUrl === void 0 ? void 0 : callBackUrl.url) === null || _a === void 0 ? void 0 : _a.includes(platformCallbackUrl))) ||
                (semver_1.default.gte(this.hw.fw_version, '1.4.0') &&
                    !((_b = callBackUrl === null || callBackUrl === void 0 ? void 0 : callBackUrl.generic) === null || _b === void 0 ? void 0 : _b.includes(platformCallbackUrl)))) {
                this.log.warn('Update existing callback URL ->', callBackUrl);
                this.platform.setButtonCallbackUrl({
                    baseUrl: this.baseUrl,
                    token: this.device.token,
                    oldUrl: callBackUrl.url,
                    endpoints: endpoints,
                });
            }
            else {
                this.log.debug('Callback URL already set ->', callBackUrl);
            }
        })
            .catch(this.handleRequestErrors.bind(this));
    }
    // Get updated device info and update the corresponding values
    getDeviceStateUpdate() {
        return this.getDeviceState()
            .then((state) => {
            if (typeof state !== 'undefined') {
                // Outputs
                this.dingzStates.Dimmers = state.dimmers;
                this.dingzStates.LED = state.led;
                // Sensors
                this.dingzStates.Temperature = state.sensors.room_temperature;
                this.dingzStates.Brightness = state.sensors.brightness;
                // Lamellas
                this.dingzStates.WindowCovers = state.blinds;
                if (this.reachabilityState ||
                    this.configTimestamp !== state.config.timestamp) {
                    // Push config change
                    this.log.debug('Config changes, update accessories');
                    DingzAccessory.getConfig({
                        address: this.device.address,
                        token: this.device.token,
                    })
                        .then(({ dingzDevices, dingzConfig }) => {
                        const newHw = dingzDevices[this.device.mac];
                        this.configTimestamp = state.config.timestamp;
                        if (this.mustInit) {
                            this.log.warn('Deferred initialization started');
                            this.mustInit = false;
                            this.device.hwInfo = newHw;
                            this.hw = newHw;
                            this.config = dingzConfig;
                            this.accessory.context.device = this.device;
                            this.accessory.context.config = this.config;
                            this.reconfigureAccessory(true);
                        }
                        else {
                            this.updateAccessory({
                                deviceHwInfo: newHw,
                                dingzConfig: dingzConfig,
                            });
                        }
                    })
                        .catch(this.handleRequestErrors.bind(this));
                }
                // Push the Update to HomeBridge
                this.eb.emit("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */);
                return Promise.resolve();
            }
            else {
                return Promise.reject(new errors_1.DeviceNotReachableError());
            }
        })
            .catch((e) => {
            // invalidate config
            return Promise.reject(e);
        });
    }
    addTemperatureService() {
        var _a;
        const temperatureService = (_a = this.accessory.getService(this.platform.Service.TemperatureSensor)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.TemperatureSensor);
        temperatureService.setCharacteristic(this.platform.Characteristic.Name, 'Temperature');
        // create handlers for required characteristics
        temperatureService
            .getCharacteristic(this.platform.Characteristic.CurrentTemperature)
            .on("get" /* CharacteristicEventTypes.GET */, this.getTemperature.bind(this));
        this.eb.on("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */, this.updateTemperature.bind(this, temperatureService));
    }
    updateTemperature(temperatureService) {
        const temperature = this.dingzStates.Temperature;
        if (temperature !== null || temperature !== undefined) {
            temperatureService
                .getCharacteristic(this.platform.Characteristic.CurrentTemperature)
                .updateValue(temperature);
        }
    }
    /**
     * Handle the "GET" requests from HomeKit
     * to get the current value of the "Current Temperature" characteristic
     */
    getTemperature(callback) {
        // set this to a valid value for CurrentTemperature
        const temperature = this.dingzStates.Temperature;
        this.log.debug('Get Characteristic Temperature ->', temperature, '° C');
        // Don't return a value if temperature is not defined
        if (temperature !== null && temperature !== undefined) {
            callback(this.reachabilityState, temperature);
        }
        else {
            callback(new Error('No valid Temperature value'));
        }
    }
    addLightSensorService() {
        // Add the LightSensor that's integrated in the dingz
        // API: /api/v1/light
        var _a;
        const lightService = (_a = this.accessory.getService(this.platform.Service.LightSensor)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.LightSensor);
        lightService.setCharacteristic(this.platform.Characteristic.Name, 'Light');
        this.eb.on("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */, this.updateLightSensor.bind(this, lightService));
    }
    updateLightSensor(lightService) {
        const intensity = this.dingzStates.Brightness;
        lightService
            .getCharacteristic(this.platform.Characteristic.CurrentAmbientLightLevel)
            .updateValue((0, limit_number_1.default)(0.0001, 100000, intensity)); // fixes #300
    }
    configureBlinds(initHandlers = false) {
        // This is the block for the multiple services (Dimmers 1-4 / Blinds 1-2 / Buttons 1-4)
        // If "Input" is set, Dimmer 1 won't work. We have to take this into account
        if (!this.config.windowCoveringConfig) {
            return;
        }
        const w = this.config.windowCoveringConfig.blinds;
        /** DIP Switch
         * 0			M1& M2		(2 blinds)
         * 1			1/2L & M2	(1 blind (M2) and 2 lights)
         * 2			3/4L & M1	(1 blind (M1) and 2 lights)
         * 3			1/2/3/4L		(4 lights)
         */
        switch (this.hw.dip_config) {
            case 3:
                // DIP = 0: D0, D1, D2, D3; (Subtypes) (Unless Input, then D1, D2, D3)
                // D1, D2, D3
                this.reconfigureWindowCovering({
                    id: 'M1',
                    connected: false,
                });
                this.reconfigureWindowCovering({
                    id: 'M2',
                    connected: false,
                });
                break;
            case 2:
                // DIP = 1: M1, D2, D3;
                if (w && w[0]) {
                    this.reconfigureWindowCovering({
                        name: w[0].name && w[0].name !== '' ? w[0].name : 'Blind 1',
                        id: 'M1',
                        index: 0,
                        connected: true,
                        initHandlers: initHandlers,
                    });
                }
                this.reconfigureWindowCovering({
                    id: 'M2',
                    connected: false,
                });
                // Dimmers are always 0 based
                // i.e. if outputs 1 / 2 are for blinds, outputs 3/4 will be dimmer 0/1
                // We use the "index" value of the dingz to determine what to use
                break;
            case 1:
                // DIP = 2: D0, D1, M2; (Unless Input, then D1, M2);
                this.reconfigureWindowCovering({
                    id: 'M1',
                    connected: false,
                });
                if (w && w[1]) {
                    // in this configuration, the second motor has the name we need
                    this.reconfigureWindowCovering({
                        name: w[1].name && w[1].name !== '' ? w[1].name : 'Blind 2',
                        id: 'M2',
                        index: 1,
                        connected: true,
                        initHandlers: initHandlers,
                    });
                }
                break;
            case 0:
                // DIP = 3: M1, M2;
                if (w && w[0] && w[1]) {
                    this.reconfigureWindowCovering({
                        name: w[0].name && w[0].name !== '' ? w[0].name : 'Blind 1',
                        id: 'M1',
                        index: 0,
                        connected: true,
                        initHandlers: initHandlers,
                    });
                    this.reconfigureWindowCovering({
                        name: w[1].name && w[1].name !== '' ? w[1].name : 'Blind 2',
                        id: 'M2',
                        index: 1,
                        connected: true,
                        initHandlers: initHandlers,
                    });
                }
                break;
            default:
                break;
        }
    }
    configureButtons(initHandler = false) {
        // Create Buttons
        const b = this.config.buttonConfig.buttons;
        this.reconfigureButtonService({
            name: b[0].name && b[0].name !== '' ? b[0].name : 'Button 1',
            button: commonTypes_1.Module.BTN1,
            initHandler,
        });
        this.reconfigureButtonService({
            name: b[1].name && b[1].name !== '' ? b[1].name : 'Button 2',
            button: commonTypes_1.Module.BTN2,
            initHandler,
        });
        this.reconfigureButtonService({
            name: b[2].name && b[2].name !== '' ? b[2].name : 'Button 3',
            button: commonTypes_1.Module.BTN3,
            initHandler,
        });
        this.reconfigureButtonService({
            name: b[3].name && b[3].name !== '' ? b[3].name : 'Button 4',
            button: commonTypes_1.Module.BTN4,
            initHandler,
        });
        // Add Event Listeners
        this.platform.eb.on("deviceAction" /* PlatformEvent.ACTION */, (mac, action, module) => {
            var _a;
            if (mac === this.device.mac && module) {
                this.log.debug(`Module ${module} triggered -> ${action}, MAC: ${mac} (This: ${this.device.mac})`);
                switch (module) {
                    case commonTypes_1.Module.INPUT:
                        // Fix for v1.2.x firmware (#318)
                        // INPUT (Unassigned)
                        this.log.info(`Button ${module} Input -> ${action} (no action)`);
                        break;
                    case commonTypes_1.Module.PIR:
                        // PUSH MOTION
                        this.log.info(`Module ${module} Motion -> ${action}`);
                        this.log.debug('Motion Update from CALLBACK');
                        (_a = this.motionService) === null || _a === void 0 ? void 0 : _a.getCharacteristic(this.platform.Characteristic.MotionDetected).setValue(action === commonTypes_1.ButtonAction.PIR_MOTION_START ? true : false);
                        break;
                    case commonTypes_1.Module.BTN1:
                    case commonTypes_1.Module.BTN2:
                    case commonTypes_1.Module.BTN3:
                    case commonTypes_1.Module.BTN4:
                        {
                            this.dingzStates.Buttons[module].event = action !== null && action !== void 0 ? action : 1;
                            this.dingzStates.Buttons[module].state =
                                this.dingzStates.Buttons[module].state === dingzTypes_1.ButtonState.OFF
                                    ? dingzTypes_1.ButtonState.ON
                                    : dingzTypes_1.ButtonState.OFF;
                            const service = this.accessory.getServiceById(this.platform.Service.StatelessProgrammableSwitch, module);
                            const ProgrammableSwitchEvent = this.platform.Characteristic.ProgrammableSwitchEvent;
                            service === null || service === void 0 ? void 0 : service.getCharacteristic(this.platform.Characteristic.ProgrammableSwitchOutputState).updateValue(this.dingzStates.Buttons[module].state);
                            this.log.info(`Button ${module} (${service === null || service === void 0 ? void 0 : service.displayName}) pressed -> ${action}`);
                            switch (action) {
                                case commonTypes_1.ButtonAction.SINGLE_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.SINGLE_PRESS);
                                    break;
                                case commonTypes_1.ButtonAction.DOUBLE_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.DOUBLE_PRESS);
                                    break;
                                case commonTypes_1.ButtonAction.LONG_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.LONG_PRESS);
                                    break;
                            }
                            // Immediately update states after button pressed
                            this.getDeviceStateUpdate();
                        }
                        break;
                    default:
                        this.log.error(`Unknown Module ${module} triggered -> MAC: ${mac} (This: ${this.device.mac})`);
                        break;
                }
            }
        });
    }
    reconfigureButtonService({ name, button, initHandler = false, }) {
        var _a;
        this.log.info('Adding Button Service ->', name, ' -> ', button);
        const buttonService = (_a = this.accessory.getServiceById(this.platform.Service.StatelessProgrammableSwitch, button)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.StatelessProgrammableSwitch, name !== null && name !== void 0 ? name : `Button ${button}`, // Name Dimmers according to WebUI, not API info
        button);
        buttonService.setCharacteristic(this.platform.Characteristic.Name, name !== null && name !== void 0 ? name : `Button ${button}`);
        buttonService.setCharacteristic(this.platform.Characteristic.ServiceLabelIndex, button);
        // Stateful Programmable Switches are not anymore exposed in HomeKit. However,
        //  the "ProgrammableSwitchOutputState" Characteristic added to a
        // StatelessProgrammableSwitch (i.e., a button), can be read out -- and used --
        // by third-party apps for HomeKite, allowing users to create automations
        // not only based on the button events, but also based on a state that's toggled
        if (initHandler) {
            buttonService
                .getCharacteristic(this.platform.Characteristic.ProgrammableSwitchOutputState)
                .on("get" /* CharacteristicEventTypes.GET */, this.getSwitchButtonState.bind(this, button));
            buttonService
                .getCharacteristic(this.platform.Characteristic.ProgrammableSwitchEvent)
                .on("get" /* CharacteristicEventTypes.GET */, this.getButtonState.bind(this, button));
        }
        return buttonService;
    }
    getButtonState(button, callback) {
        const currentState = this.dingzStates.Buttons[button].event;
        callback(this.reachabilityState, currentState);
    }
    getSwitchButtonState(button, callback) {
        const currentState = this.dingzStates.Buttons[button].state;
        this.log.info('Get Switch State of ->', button, '-> state:', currentState);
        callback(this.reachabilityState, currentState);
    }
    reconfigureOutput({ name, output, id, index, initHandlers, }) {
        if (output === 'not_connected') {
            this.removeOutput(id);
            return;
        }
        this.log.info(`setDimmerConfig() -> Configuring output '${name}' as '${output}' (${id}/${index})`);
        const existing = this.accessory.getServiceById(this.platform.Service.Lightbulb, id);
        const service = existing !== null && existing !== void 0 ? existing : this.accessory.addService(this.platform.Service.Lightbulb, name, // Name Dimmers according to WebUI, not API info
        id);
        // Update name
        service.getCharacteristic(this.platform.Characteristic.Name).setValue(name);
        // add / configure Brightness
        if (output === 'non_dimmable' &&
            service.testCharacteristic(this.platform.Characteristic.Brightness)) {
            service.removeCharacteristic(service.getCharacteristic(this.platform.Characteristic.Brightness));
        }
        else if (output !== 'non_dimmable') {
            try {
                // Add listener in any case, but first destroy existing ones (should fix #256)
                service.removeCharacteristic(service.getCharacteristic(this.platform.Characteristic.Brightness));
            }
            catch (e) {
                this.log.warn('Attempt to remove "Brightness" characteristic failed');
            }
            finally {
                service.addCharacteristic(this.platform.Characteristic.Brightness);
                service
                    .getCharacteristic(this.platform.Characteristic.Brightness)
                    .on("set" /* CharacteristicEventTypes.SET */, this.setBrightness.bind(this, index)); // SET - bind to the 'setBrightness` method below
            }
        }
        if (initHandlers || !existing) {
            this.setOutputHandlers(service, index);
        }
        return service;
    }
    updateDimmerState(service, index) {
        if (service && index !== null) {
            // index set
            const state = this.dingzStates.Dimmers[index];
            // Check that "state" is valid
            if (state) {
                if (service.testCharacteristic(this.platform.Characteristic.Brightness)) {
                    service
                        .getCharacteristic(this.platform.Characteristic.Brightness)
                        .updateValue(state.output);
                }
                service
                    .getCharacteristic(this.platform.Characteristic.On)
                    .updateValue(state.on);
            }
        }
    }
    setOn(index, value, callback) {
        this.dingzStates.Dimmers[index].on = value;
        this.setDeviceDimmer(index, callback, value);
    }
    /**
     * Handle the "GET" requests from HomeKit
     */
    getOn(index, callback) {
        var _a;
        const isOn = (_a = this.dingzStates.Dimmers[index]) === null || _a === void 0 ? void 0 : _a.on;
        this.log.debug('Get Characteristic On ->', isOn, '');
        if (isOn !== null && isOn !== undefined && typeof isOn === 'boolean') {
            callback(this.reachabilityState, isOn);
        }
        else {
            callback(new Error('No temperature value available'));
        }
    }
    async setBrightness(index, value, callback) {
        const isOn = value > 0 ? true : false;
        this.dingzStates.Dimmers[index].output = value;
        this.dingzStates.Dimmers[index].on = isOn;
        this.setDeviceDimmer(index, callback, isOn, value);
    }
    /**
     * Configure a WindowCovering
     */
    reconfigureWindowCovering({ name, id, index, connected, initHandlers = false, }) {
        const existing = this.accessory.getServiceById(this.platform.Service.WindowCovering, id);
        if (connected && name && index !== undefined) {
            this.log.info(`configureWindowCoveringService() -> add Blind '${name}' (${id}/${index})`);
            const service = existing !== null && existing !== void 0 ? existing : this.accessory.addService(this.platform.Service.WindowCovering, name, id);
            // each service must implement at-minimum the "required characteristics" for the given service type
            // see https://developers.homebridge.io/#/service/Lightbulb
            service
                .getCharacteristic(this.platform.Characteristic.Name)
                .setValue(name);
            if (initHandlers || !existing) {
                this.initWindowCoveringService(service, index);
            }
        }
        else if (!connected && existing) {
            this.log.info(`configureWindowCoveringService() -> remove Blind (${id}/${index})`);
            this.accessory.removeService(existing);
        }
    }
    // Add WindowCovering (Blinds)
    initWindowCoveringService(service, index) {
        // register handlers for the WindoCovering TargetPosition
        service
            .getCharacteristic(this.platform.Characteristic.TargetPosition)
            .on("set" /* CharacteristicEventTypes.SET */, this.setPosition.bind(this, index));
        // Set min/max Values
        // FIXME: Implement different lamella/blind modes #24
        service
            .getCharacteristic(this.platform.Characteristic.TargetHorizontalTiltAngle)
            .setProps({
            minValue: 0,
            maxValue: 90,
            minStep: this.platform.config.minStepTiltAngle,
        }) // dingz Maximum values
            .on("set" /* CharacteristicEventTypes.SET */, this.setTiltAngle.bind(this, index));
        service
            .getCharacteristic(this.platform.Characteristic.CurrentPosition)
            .on("get" /* CharacteristicEventTypes.GET */, this.getPosition.bind(this, index));
        service
            .getCharacteristic(this.platform.Characteristic.CurrentHorizontalTiltAngle)
            .on("get" /* CharacteristicEventTypes.GET */, this.getTiltAngle.bind(this, index));
        service
            .getCharacteristic(this.platform.Characteristic.PositionState)
            .on("get" /* CharacteristicEventTypes.GET */, this.getPositionState.bind(this, index));
        // Subscribe to the update event
        this.eb.on("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */, this.updateWindowCoveringState.bind(this, index, service));
    }
    // Window Covering functions
    updateWindowCoveringState(index, service) {
        const id = this.getWindowCoveringId(index);
        const state = this.dingzStates.WindowCovers[id];
        if (state) {
            /**
             * TODO: Fix Hardware Buttons and UI buttons
             * It can be complicated:
             * - We're moving by setting new positions in the UI [x]
             * - We're moving by pressing the "up/down" buttons in the UI or Hardware [x]
             */
            service
                .getCharacteristic(this.platform.Characteristic.TargetPosition)
                .updateValue(state.position);
            service
                .getCharacteristic(this.platform.Characteristic.TargetHorizontalTiltAngle)
                .updateValue((state.lamella / 100) * 90); // Lamella position set in ° in HomeKit
            let positionState;
            switch (state.moving) {
                case 'down':
                    positionState = this.platform.Characteristic.PositionState.DECREASING;
                    break;
                case 'up':
                    positionState = this.platform.Characteristic.PositionState.INCREASING;
                    break;
                case 'stop':
                    positionState = this.platform.Characteristic.PositionState.STOPPED;
                    service
                        .getCharacteristic(this.platform.Characteristic.CurrentPosition)
                        .updateValue(state.position);
                    service
                        .getCharacteristic(this.platform.Characteristic.CurrentHorizontalTiltAngle)
                        .updateValue((state.lamella / 100) * 90); // Lamella position set in ° in HomeKit
                    break;
            }
            service
                .getCharacteristic(this.platform.Characteristic.PositionState)
                .updateValue(positionState);
        }
    }
    async setPosition(index, position, callback) {
        const id = this.getWindowCoveringId(index);
        const windowCovering = this.dingzStates.WindowCovers[id];
        if (windowCovering) {
            // Make sure we're setting motion when changing the position
            if (position > windowCovering.position) {
                windowCovering.moving = 'up';
            }
            else if (position < windowCovering.position) {
                windowCovering.moving = 'down';
            }
            else {
                windowCovering.moving = 'stop';
            }
            this.log.warn('Blinds moving: ', windowCovering.moving, '-->', position);
            await this.setWindowCovering({
                id: id,
                blind: position,
                lamella: (windowCovering.lamella / 90) * 100,
                callback: callback,
            });
        }
    }
    getPosition(index, callback) {
        var _a, _b;
        this.log.debug('WindowCoverings: ', JSON.stringify(this.dingzStates.WindowCovers));
        const id = this.getWindowCoveringId(index);
        const blind = (_b = (_a = this.dingzStates.WindowCovers[id]) === null || _a === void 0 ? void 0 : _a.position) !== null && _b !== void 0 ? _b : 0; // fixes #300 - down by default
        this.log.debug('Get Characteristic for WindowCovering', index, 'Current Position ->', blind);
        callback(this.reachabilityState, blind);
    }
    async setTiltAngle(index, angle, callback) {
        this.log.debug('Set Characteristic TargetHorizontalTiltAngle on ', index, '->', `${angle}°`);
        const id = this.getWindowCoveringId(index);
        if (this.dingzStates.WindowCovers[id]) {
            await this.setWindowCovering({
                id: id,
                blind: this.dingzStates.WindowCovers[id].position,
                lamella: (angle / 90) * 100,
                callback: callback,
            });
        }
    }
    /**
     * Handle the "GET" requests from HomeKit
     */
    getTiltAngle(index, callback) {
        var _a, _b;
        this.log.debug('WindowCoverings: ', JSON.stringify(this.dingzStates.WindowCovers));
        const id = this.getWindowCoveringId(index);
        const tiltAngle = (_b = (_a = this.dingzStates.WindowCovers[id]) === null || _a === void 0 ? void 0 : _a.lamella) !== null && _b !== void 0 ? _b : 0; // Implement fix for #300s
        this.log.debug('Get Characteristic for WindowCovering', index, 'Current TiltAngle ->', tiltAngle);
        // FIXES #371, #419: internally, it's % (but only in newer firmware, v1.2.0 and lower has ° as well), HomeKit expects °
        const maxTiltValue = semver_1.default.lt(this.hw.fw_version, '1.2.0') ? 90 : 100;
        callback(this.reachabilityState, (tiltAngle / maxTiltValue) * 90);
    }
    getPositionState(index, callback) {
        var _a;
        this.log.debug('WindowCoverings: ', JSON.stringify(this.dingzStates.WindowCovers));
        let positionState = this.platform.Characteristic.PositionState.STOPPED;
        const id = this.getWindowCoveringId(index);
        const moving = (_a = this.dingzStates.WindowCovers[id]) === null || _a === void 0 ? void 0 : _a.moving;
        if (moving) {
            switch (moving) {
                case 'down':
                    positionState = this.platform.Characteristic.PositionState.DECREASING;
                    break;
                case 'up':
                    positionState = this.platform.Characteristic.PositionState.INCREASING;
                    break;
                case 'stop':
                    positionState = this.platform.Characteristic.PositionState.STOPPED;
                    break;
            }
            this.log.debug('Get Characteristic for WindowCovering', index, 'Current Position State ->', positionState);
        }
        callback(this.reachabilityState, positionState);
    }
    /**
     * Motion Service Methods
     */
    addMotionService() {
        var _a, _b;
        this.log.info('addMotionService() -> adding motion service');
        this.motionService =
            (_a = this.accessory.getService(this.platform.Service.MotionSensor)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.MotionSensor);
        this.motionService.setCharacteristic(this.platform.Characteristic.Name, 'Motion');
        // Only check for motion if we have a PIR and set the Interval
        if ((_b = this.platform.config.motionPoller) !== null && _b !== void 0 ? _b : true) {
            this.log.info('Motion POLLING enabled');
            const motionInterval = setInterval(() => {
                this.getDeviceMotion()
                    .then((data) => {
                    var _a;
                    if (data === null || data === void 0 ? void 0 : data.success) {
                        const isMotion = data.motion;
                        // Only update if motionService exists *and* if there's a change in motion'
                        if (this.motionService && this.dingzStates.Motion !== isMotion) {
                            this.log.debug('Motion Update from POLLER');
                            this.dingzStates.Motion = isMotion;
                            (_a = this.motionService) === null || _a === void 0 ? void 0 : _a.getCharacteristic(this.platform.Characteristic.MotionDetected).updateValue(isMotion);
                        }
                    }
                    else {
                        throw new errors_1.DeviceNotReachableError(`Device can not be reached -> ${this.accessory.displayName}-> ${this.device.address}`);
                    }
                })
                    .catch((e) => {
                    this.log.error('Error -> unable to fetch DeviceMotion data', e.name, e.toString());
                });
            }, 2000); // Shorter term updates for motion sensor
            this.motionTimer = motionInterval;
        }
    }
    // Remove motion service
    removeMotionService() {
        // Remove motionService & motionTimer
        if (this.motionTimer) {
            clearTimeout(this.motionTimer);
            this.motionTimer = undefined;
        }
        const service = this.accessory.getService(this.platform.Service.MotionSensor);
        if (service) {
            this.log.info('removeMotionService(): Removing Motion service ->', service.displayName);
            this.accessory.removeService(service);
        }
    }
    // Updates the Accessory (e.g. if the config has changed)
    updateAccessory({ deviceHwInfo, dingzConfig, }) {
        if (this.reachabilityState !== null) {
            this.log.warn('Device recovered from unreachable state');
            this.reachabilityState = null;
        }
        this.log.warn(`Config changed -> will update accessory (${this.device.address})`);
        let updatedDingzDeviceInfo;
        try {
            updatedDingzDeviceInfo = deviceHwInfo !== null && deviceHwInfo !== void 0 ? deviceHwInfo : this.hw;
            if (this.hw && this.hw.has_pir !== updatedDingzDeviceInfo.has_pir) {
                // Update PIR Service
                this.log.warn('Update accessory -> PIR config changed.');
                if (updatedDingzDeviceInfo.has_pir) {
                    // Add PIR service
                    this.addMotionService();
                }
                else {
                    // Remove PIR service
                    this.removeMotionService();
                }
            }
            // Update output, blind, input services
            this.device.hwInfo = deviceHwInfo;
            this.hw = deviceHwInfo;
            this.config = dingzConfig;
            this.accessory.context.device = this.device;
            this.accessory.context.config = this.config;
            this.configureOutputs();
            this.configureBlinds();
            this.configureButtons();
        }
        finally {
            if (updatedDingzDeviceInfo) {
                this.accessory.context.device.dingzDeviceInfo = updatedDingzDeviceInfo;
            }
        }
    }
    /**
     * Updates the dimemr services based on their current configuration
     */
    configureOutputs(initHandlers = false) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        // Figure out what we have here
        if (!this.config || (this.config && !this.config.dimmerConfig)) {
            return;
        }
        const d = this.config.dimmerConfig.dimmers;
        const i = this.config.inputConfig.inputs;
        switch (this.hw.dip_config) {
            case 3:
                if (i[0] && !i[0].active && d[0].output !== 'not_connected') {
                    this.reconfigureOutput({
                        name: (_a = d[0].name) !== null && _a !== void 0 ? _a : 'Output 1',
                        id: 'D1',
                        index: 0,
                        output: d[0].output,
                        initHandlers: initHandlers,
                    });
                }
                else {
                    this.removeOutput('D1');
                }
                this.reconfigureOutput({
                    name: (_b = d[1].name) !== null && _b !== void 0 ? _b : 'Output 2',
                    id: 'D2',
                    index: 1,
                    output: d[1].output,
                    initHandlers: initHandlers,
                });
                this.reconfigureOutput({
                    name: (_c = d[2].name) !== null && _c !== void 0 ? _c : 'Output 3',
                    id: 'D3',
                    index: 2,
                    output: d[2].output,
                    initHandlers: initHandlers,
                });
                this.reconfigureOutput({
                    name: (_d = d[3].name) !== null && _d !== void 0 ? _d : 'Output 4',
                    id: 'D4',
                    index: 3,
                    output: d[3].output,
                    initHandlers: initHandlers,
                });
                break;
            case 2:
                this.removeOutput('D1');
                this.removeOutput('D2');
                this.reconfigureOutput({
                    name: (_e = d[2].name) !== null && _e !== void 0 ? _e : 'Output 3',
                    id: 'D3',
                    index: 0,
                    output: d[2].output,
                    initHandlers: initHandlers,
                });
                this.reconfigureOutput({
                    name: (_f = d[3].name) !== null && _f !== void 0 ? _f : 'Output 4',
                    id: 'D4',
                    index: 1,
                    output: d[3].output,
                    initHandlers: initHandlers,
                });
                break;
            case 1:
                if (i[0] && !i[0].active && d[0].output !== 'not_connected') {
                    this.reconfigureOutput({
                        name: (_g = d[0].name) !== null && _g !== void 0 ? _g : 'Output 1',
                        id: 'D1',
                        index: 0,
                        output: d[0].output,
                        initHandlers: initHandlers,
                    });
                }
                else {
                    this.removeOutput('D1');
                }
                this.reconfigureOutput({
                    name: (_h = d[1].name) !== null && _h !== void 0 ? _h : 'Output 2',
                    id: 'D2',
                    index: 1,
                    output: d[1].output,
                    initHandlers: initHandlers,
                });
                this.removeOutput('D3');
                this.removeOutput('D4');
                break;
            case 0:
                this.removeOutput('D1');
                this.removeOutput('D2');
                this.removeOutput('D3');
                this.removeOutput('D4');
                break;
            default:
                break;
        }
    }
    removeOutput(id) {
        const service = this.accessory.getServiceById(this.platform.Service.Lightbulb, id);
        if (service) {
            // Remove dimmer since not connected
            this.log.info(`setDimmerConfig() -> Removing output (${id})`);
            this.accessory.removeService(service);
            return;
        }
    }
    addLEDService() {
        var _a;
        const systemConfig = this.config.systemConfig;
        // get the LightBulb service if it exists, otherwise create a new LightBulb service
        // you can create multiple services for each accessory
        const ledService = (_a = this.accessory.getServiceById(this.platform.Service.Lightbulb, 'LED')) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.Lightbulb, 'LED', 'LED');
        // set the service name, this is what is displayed as the default name on the Home app
        let ledName = 'LED';
        if (systemConfig === null || systemConfig === void 0 ? void 0 : systemConfig.dingz_name) {
            ledName = `${systemConfig === null || systemConfig === void 0 ? void 0 : systemConfig.dingz_name} LED`;
        }
        ledService.setCharacteristic(this.platform.Characteristic.Name, ledName);
        // register handlers for the On/Off Characteristic
        ledService
            .getCharacteristic(this.platform.Characteristic.On)
            .on("set" /* CharacteristicEventTypes.SET */, this.setLEDOn.bind(this)) // SET - bind to the `setOn` method below
            .on("get" /* CharacteristicEventTypes.GET */, this.getLEDOn.bind(this)); // GET - bind to the `getOn` method below
        // register handlers for the Brightness Characteristic
        ledService
            .getCharacteristic(this.platform.Characteristic.Brightness)
            .on("set" /* CharacteristicEventTypes.SET */, this.setLEDBrightness.bind(this)); // SET - bind to the 'setBrightness` method below
        // register handlers for the Brightness Characteristic
        ledService
            .getCharacteristic(this.platform.Characteristic.Hue)
            .on("set" /* CharacteristicEventTypes.SET */, this.setLEDHue.bind(this)); // SET - bind to the 'setBrightness` method below
        // register handlers for the Brightness Characteristic
        ledService
            .getCharacteristic(this.platform.Characteristic.Saturation)
            .on("set" /* CharacteristicEventTypes.SET */, this.setLEDSaturation.bind(this)); // SET - bind to the 'setBrightness` method below
        this.eb.on("pushStateUpdate" /* AccessoryEvent.PUSH_STATE_UPDATE */, this.updateLEDState.bind(this, ledService));
    }
    updateLEDState(ledService) {
        const state = this.dingzStates.LED;
        if (state.mode === 'hsv') {
            const hsv = state.hsv.split(';');
            this.dingzStates.LED.hue = parseInt(hsv[0]);
            this.dingzStates.LED.saturation = parseInt(hsv[1]);
            this.dingzStates.LED.value = parseInt(hsv[2]);
        }
        else {
            // rgbw
            const hsv = new simple_color_converter_1.default({
                color: `hex #${state.rgb}`,
                to: 'hsv',
            });
            this.dingzStates.LED.hue = hsv.color.h;
            this.dingzStates.LED.saturation = hsv.color.s;
            this.dingzStates.LED.value = hsv.color.v;
        }
        ledService
            .getCharacteristic(this.platform.Characteristic.Hue)
            .updateValue(this.dingzStates.LED.hue);
        ledService
            .getCharacteristic(this.platform.Characteristic.Saturation)
            .updateValue(this.dingzStates.LED.saturation);
        ledService
            .getCharacteristic(this.platform.Characteristic.Brightness)
            .updateValue(this.dingzStates.LED.value);
        if (this.dingzStates.LED.on !== null &&
            this.dingzStates.LED.on !== undefined) {
            ledService
                .getCharacteristic(this.platform.Characteristic.On)
                .updateValue(this.dingzStates.LED.on);
        }
    }
    setLEDOn(value, callback) {
        this.dingzStates.LED.on = value;
        this.setDeviceLED(callback);
    }
    getLEDOn(callback) {
        const isOn = this.dingzStates.LED.on;
        if (isOn !== null && isOn !== undefined) {
            callback(this.reachabilityState, isOn);
        }
        else {
            callback(new Error('No LED status available'));
        }
    }
    setLEDHue(value, callback) {
        this.dingzStates.LED.hue = value;
        this.setDeviceLED(callback);
    }
    setLEDSaturation(value, callback) {
        this.dingzStates.LED.saturation = value;
        this.setDeviceLED(callback);
    }
    setLEDBrightness(value, callback) {
        this.dingzStates.LED.value = value;
        this.setDeviceLED(callback);
    }
    // Get Input & Dimmer Config
    static async getConfig({ address, token, }) {
        const deviceInfoEndpoint = '/api/v1/device';
        const deviceConfigEndpoint = '/api/v1/system_config';
        const inputConfigEndpoint = '/api/v1/input_config';
        const dimmerConfigEndpoint = '/api/v1/dimmer_config';
        const blindConfigEndpoint = '/api/v1/blind_config';
        const buttonConfigEndpoint = '/api/v1/button_config';
        const config = {
            baseURL: `http://${address}`,
            timeout: settings_1.RETRY_TIMEOUT * 1500,
            headers: { Token: token !== null && token !== void 0 ? token : '' },
        };
        const [dingzDevicesResponse, systemConfigResponse, inputConfigResponse, dimmerConfigResponse, blindConfigResponse, buttonConfigResponse,] = await Promise.all([
            DingzAccessory.axios.get(deviceInfoEndpoint, config),
            DingzAccessory.axios.get(deviceConfigEndpoint, config),
            DingzAccessory.axios.get(inputConfigEndpoint, config),
            DingzAccessory.axios.get(dimmerConfigEndpoint, config),
            DingzAccessory.axios.get(blindConfigEndpoint, config),
            DingzAccessory.axios.get(buttonConfigEndpoint, config),
        ]).catch((e) => {
            var _a;
            if (e.code === 'ECONNABORTED') {
                throw new errors_1.DeviceNotReachableError(`${(_a = e.code) !== null && _a !== void 0 ? _a : ''}: Device ${address} not reachable after ${settings_1.RETRY_TIMEOUT} s\n\n${e.stack}`);
            }
            else {
                throw e;
            }
        });
        const deviceConfig = {
            systemConfig: systemConfigResponse.data,
            inputConfig: inputConfigResponse.data,
            dimmerConfig: dimmerConfigResponse.data,
            windowCoveringConfig: blindConfigResponse.data,
            buttonConfig: buttonConfigResponse.data,
        };
        return {
            dingzDevices: dingzDevicesResponse.data,
            dingzConfig: deviceConfig,
        };
    }
    async getDeviceMotion() {
        const getMotionEndpoint = '/api/v1/motion';
        return await this.request
            .get(getMotionEndpoint)
            .then((response) => {
            return response.data;
        })
            .catch(this.handleRequestErrors.bind(this));
    }
    // Set individual dimmer
    async setDeviceDimmer(index, callback, isOn, level) {
        // /api/v1/dimmer/<DIMMER>/on/?value=<value>
        const setDimmerEndpoint = `/api/v1/dimmer/${index}/${isOn ? 'on' : 'off'}/${level ? '?value=' + level : ''}`;
        await this.request
            .post(setDimmerEndpoint)
            .catch(this.handleRequestErrors.bind(this))
            .finally(() => {
            callback(this.reachabilityState);
        });
    }
    // Set individual dimmer
    setWindowCovering({ id, blind, lamella, callback, }) {
        // The API only accepts integer numbers.
        // As we juggle with ° vs %, we must round
        // the values for blind and lamella to the nearest integer
        blind = Math.round(blind);
        lamella = Math.round(lamella);
        this.log.debug(`Setting WindowCovering ${id} to position ${blind} and angle ${lamella}°`);
        // The API says the parameters can be omitted. This is not true
        // {{ip}}/api/v1/shade/0?blind=<value>&lamella=<value>
        const setWindowCoveringEndpoint = `${this.baseUrl}/api/v1/shade/${id}`;
        this.request
            .post(setWindowCoveringEndpoint, qs_1.default.stringify({
            blind: blind,
            lamella: lamella,
        }, { encode: false }))
            .catch(this.handleRequestErrors.bind(this))
            .finally(() => {
            callback(this.reachabilityState);
        });
    }
    /**
     * Set the LED on the dingz
     * @param callback: Characteristic callback
     */
    setDeviceLED(callback) {
        const state = this.dingzStates.LED;
        const color = `${state.hue};${state.saturation};${state.value}`;
        const setLEDEndpoint = `${this.baseUrl}/api/v1/led/set`;
        this.request
            .post(setLEDEndpoint, qs_1.default.stringify({
            action: state.on ? 'on' : 'off',
            color: color,
            mode: 'hsv',
            ramp: 150,
        }, { encode: false }))
            .catch(this.handleRequestErrors.bind(this))
            .finally(() => {
            callback(this.reachabilityState);
        });
    }
    // Get the current state
    // This function is called regularly and contains all necessary
    // information for an update of all sensors and states
    async getDeviceState() {
        const getDeviceStateEndpoint = `${this.baseUrl}/api/v1/state`;
        return await this.request.get(getDeviceStateEndpoint).then((response) => {
            return response.data;
        });
    }
    /**
     * Returns the callback URL for the device
     */
    async getButtonCallbackUrl() {
        // FIXES #511: different endpoint URLs for Callback from FW v1.4.x forward
        const getCallbackEndpoint = semver_1.default.gte(this.hw.fw_version, '1.4.0')
            ? '/api/v1/action/generic'
            : '/api/v1/action/generic/generic';
        this.log.debug('Getting the callback URL -> ', getCallbackEndpoint);
        return await this.request.get(getCallbackEndpoint).then((response) => {
            return response.data;
        });
    }
    /**
     * Get the right blind state.
     * It's complicated ...
     * @param index *aboslute* index value (i.e. 0/1)
     */
    getWindowCoveringId(index) {
        if (index === 0) {
            // we have either two blinds *but* index 0
            return 0;
        }
        else if (this.dingzStates.WindowCovers.length === 1 && index === 1) {
            // we have one blind and it's index 1
            return 0;
        }
        else {
            return 1;
        }
    }
    /**
     * Enable the PIR callback
     */
    enablePIRCallback() {
        const setActionEndpoint = '/api/v1/action/pir/press_release/enable';
        this.log.debug('Enabling the PIR callback -> ');
        this.request
            .post(setActionEndpoint)
            .catch(this.handleRequestErrors.bind(this));
    }
}
exports.DingzAccessory = DingzAccessory;
//# sourceMappingURL=dingzAccessory.js.map