"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MyStromButtonPlusAccessory = void 0;
const cockatiel_1 = require("cockatiel");
const myStromTypes_1 = require("./lib/myStromTypes");
const commonTypes_1 = require("./lib/commonTypes");
const dingzTypes_1 = require("./lib/dingzTypes");
const myStromButtonAccessory_1 = require("./myStromButtonAccessory");
// Policy for long running tasks, retry every hour
const retrySlow = cockatiel_1.Policy.handleAll()
    .orWhenResult((retry) => retry === true)
    .retry()
    .exponential({ initialDelay: 10000, maxDelay: 60 * 60 * 1000 });
/**
 * Platform Accessory
 * An instance of this class is created for each accessory your platform registers
 * Each accessory may expose multiple services of different service types.
 */
class MyStromButtonPlusAccessory extends myStromButtonAccessory_1.MyStromButtonAccessory {
    constructor(_platform, _accessory) {
        var _a, _b;
        super(_platform, _accessory);
        this._platform = _platform;
        this._accessory = _accessory;
        // Eventually replaced by:
        // FIXME: Upgrade for multiple buttons
        this.buttonActions = [
            commonTypes_1.ButtonAction.SINGLE_PRESS,
            commonTypes_1.ButtonAction.SINGLE_PRESS,
            commonTypes_1.ButtonAction.SINGLE_PRESS,
            commonTypes_1.ButtonAction.SINGLE_PRESS,
        ];
        this.buttonStates = [
            dingzTypes_1.ButtonState.OFF,
            dingzTypes_1.ButtonState.OFF,
            dingzTypes_1.ButtonState.OFF,
            dingzTypes_1.ButtonState.OFF,
        ];
        this.buttonPlusState = {
            temperature: 0,
            humidity: 0,
            battery: {
                voltage: 0,
                charging: false,
            },
            charger: {
                voltage: 0,
                connected: false,
            },
        };
        // Set Base URL
        this.log.debug('Setting informationService Characteristics ->', this.device.model);
        this.accessory
            .getService(this.platform.Service.AccessoryInformation)
            .setCharacteristic(this.platform.Characteristic.Manufacturer, 'MyStrom AG')
            .setCharacteristic(this.platform.Characteristic.AppMatchingIdentifier, 'ch.mystrom.iOSApp')
            .setCharacteristic(this.platform.Characteristic.Model, this.device.model)
            .setCharacteristic(this.platform.Characteristic.FirmwareRevision, 'N/A')
            .setCharacteristic(this.platform.Characteristic.HardwareRevision, 'PQWBB1')
            .setCharacteristic(this.platform.Characteristic.SerialNumber, this.device.mac);
        // get the StatelessProgrammableSwitch service if it exists,
        // otherwise create a new StatelessProgrammableSwitch service
        // you can create multiple services for each accessory
        this.log.debug('Create Stateless Programmable Switch');
        this.buttonServiceSetup();
        // Button Plus 2nd Gen has a temperature sensor, make it available here
        // create a new Temperature Sensor service
        const temperatureService = (_a = this.accessory.getService(this.platform.Service.TemperatureSensor)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.TemperatureSensor);
        temperatureService.setCharacteristic(this.platform.Characteristic.Name, 'Temperature');
        // create handlers for required characteristics
        temperatureService
            .getCharacteristic(this.platform.Characteristic.CurrentTemperature)
            .on("get" /* CharacteristicEventTypes.GET */, this.getTemperature.bind(this));
        // Button Plus 2nd Gen has a humidity sensor, make it available here
        // create a new Humidity Sensor service
        const humidityService = (_b = this.accessory.getService(this.platform.Service.HumiditySensor)) !== null && _b !== void 0 ? _b : this.accessory.addService(this.platform.Service.HumiditySensor);
        humidityService.setCharacteristic(this.platform.Characteristic.Name, 'Humidity');
        // create handlers for required characteristics
        humidityService
            .getCharacteristic(this.platform.Characteristic.CurrentRelativeHumidity)
            .on("get" /* CharacteristicEventTypes.GET */, this.getHumidity.bind(this));
        // Set the callback URL (Override!)
        retrySlow.execute(() => {
            this.platform.setButtonCallbackUrl({
                baseUrl: this.baseUrl,
                token: this.device.token,
                endpoints: ['generic/generic'], // Buttons need the 'generic' endpoint specifically set
            });
        });
    }
    getTemperature(callback) {
        var _a;
        const temperature = (_a = this.buttonPlusState) === null || _a === void 0 ? void 0 : _a.temperature;
        this.log.debug('Get Characteristic Temperature ->', temperature, '° C');
        // Don't return a value if temperature is not defined
        if (temperature !== null && temperature !== undefined) {
            callback(this.reachabilityState, temperature);
        }
        else {
            callback(new Error('No valid temperature value'));
        }
    }
    getHumidity(callback) {
        var _a;
        const humidity = (_a = this.buttonPlusState) === null || _a === void 0 ? void 0 : _a.humidity;
        this.log.debug('Get Characteristic Humidity ->', humidity, '% (relative humidity)');
        // Don't return a value if humidity is not defined
        if (humidity !== null && humidity !== undefined) {
            callback(this.reachabilityState, humidity);
        }
        else {
            callback(new Error('No valid humidity value'));
        }
    }
    // FIXME: initHandler is probably not needed
    buttonServiceSetup(initHandler = false) {
        // Create Buttons
        this.reconfigureButtonService({
            name: 'Button 1',
            button: commonTypes_1.Module.BTN1,
            initHandler,
        });
        this.reconfigureButtonService({
            name: 'Button 2',
            button: commonTypes_1.Module.BTN2,
            initHandler,
        });
        this.reconfigureButtonService({
            name: 'Button 3',
            button: commonTypes_1.Module.BTN3,
            initHandler,
        });
        this.reconfigureButtonService({
            name: 'Button 4',
            button: commonTypes_1.Module.BTN4,
            initHandler,
        });
        // Add Event Listeners
        this.platform.eb.on("deviceAction" /* PlatformEvent.ACTION */, (mac, action, module, battery, temperature, humidity) => {
            if (mac === this.device.mac && module) {
                this.log.debug(`Module ${module} triggered -> ${action}, MAC: ${mac} (This: ${this.device.mac})`, `\nBatt [V]: ${battery}, °C: ${temperature}, RH: ${humidity})`);
                // battery, temperature and humidity are reported in any case
                const batteryService = this.accessory.getService(this.platform.Service.Battery);
                if (batteryService && battery) {
                    // TODO: Check if true
                    // MyStrom Button+ (2nd Gen) reportedly returns voltages when updating via button actions
                    // see https://github.com/myStrom/mystrom-button/blob/master/user/peri.c (IQS)
                    // MAX: 4500 mv
                    // MIN: 3000 mv
                    const range = myStromTypes_1.MyStromButtonPlusBattery.BATTERY_MAX -
                        myStromTypes_1.MyStromButtonPlusBattery.BATTERY_MIN;
                    this.batteryLevel =
                        ((battery - myStromTypes_1.MyStromButtonPlusBattery.BATTERY_MIN) * 100) / range;
                    batteryService
                        .getCharacteristic(this.platform.Characteristic.BatteryLevel)
                        .updateValue(this.batteryLevel);
                    this.log.debug(`Setting battery level to ${this.batteryLevel}%`);
                }
                const temperatureService = this.accessory.getService(this.platform.Service.TemperatureSensor);
                if (temperatureService && temperature) {
                    this.buttonPlusState.temperature = temperature;
                    temperatureService
                        .getCharacteristic(this.platform.Characteristic.CurrentTemperature)
                        .updateValue(temperature);
                    this.log.debug(`Setting temperature to ${temperature}°`);
                }
                const humidityService = this.accessory.getService(this.platform.Service.HumiditySensor);
                if (humidityService && humidity) {
                    this.buttonPlusState.humidity = humidity;
                    humidityService
                        .getCharacteristic(this.platform.Characteristic.CurrentRelativeHumidity)
                        .updateValue(humidity);
                    this.log.debug(`Setting humidty level to ${humidity}%`);
                }
                switch (module) {
                    case commonTypes_1.Module.BTN1:
                    case commonTypes_1.Module.BTN2:
                    case commonTypes_1.Module.BTN3:
                    case commonTypes_1.Module.BTN4:
                        {
                            this.buttonActions[module] = action !== null && action !== void 0 ? action : 1;
                            this.buttonStates[module] =
                                this.buttonStates[module] === dingzTypes_1.ButtonState.OFF
                                    ? dingzTypes_1.ButtonState.ON
                                    : dingzTypes_1.ButtonState.OFF;
                            const service = this.accessory.getServiceById(this.platform.Service.StatelessProgrammableSwitch, module);
                            const ProgrammableSwitchEvent = this.platform.Characteristic.ProgrammableSwitchEvent;
                            service === null || service === void 0 ? void 0 : service.getCharacteristic(this.platform.Characteristic.ProgrammableSwitchOutputState).updateValue(this.buttonStates[module]);
                            this.log.info(`Button ${module} (${service === null || service === void 0 ? void 0 : service.displayName}) pressed -> ${action}`);
                            switch (action) {
                                case commonTypes_1.ButtonAction.SINGLE_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.SINGLE_PRESS);
                                    break;
                                case commonTypes_1.ButtonAction.DOUBLE_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.DOUBLE_PRESS);
                                    break;
                                case commonTypes_1.ButtonAction.LONG_PRESS:
                                    service === null || service === void 0 ? void 0 : service.getCharacteristic(ProgrammableSwitchEvent).setValue(ProgrammableSwitchEvent.LONG_PRESS);
                                    break;
                            }
                            // Immediately update states after button pressed
                            // this.getDeviceStateUpdate();
                        }
                        break;
                    default:
                        this.log.error(`Unknown Module ${module} triggered -> MAC: ${mac} (This: ${this.device.mac})`);
                        break;
                }
            }
        });
    }
    reconfigureButtonService({ name, button, initHandler = false, }) {
        var _a;
        this.log.info('Adding Button Service ->', name, ' -> ', button);
        const buttonService = (_a = this.accessory.getServiceById(this.platform.Service.StatelessProgrammableSwitch, button)) !== null && _a !== void 0 ? _a : this.accessory.addService(this.platform.Service.StatelessProgrammableSwitch, name !== null && name !== void 0 ? name : `Button ${button}`, // Name Dimmers according to WebUI, not API info
        button);
        buttonService.setCharacteristic(this.platform.Characteristic.Name, name !== null && name !== void 0 ? name : `Button ${button}`);
        buttonService.setCharacteristic(this.platform.Characteristic.ServiceLabelIndex, button);
        // Stateful Programmable Switches are not anymore exposed in HomeKit. However,
        // the "ProgrammableSwitchOutputState" Characteristic added to a
        // StatelessProgrammableSwitch (i.e., a button), can be read out -- and used --
        // by third-party apps for HomeKite, allowing users to create automations
        // not only based on the button events, but also based on a state that's toggled
        if (initHandler) {
            /* TODO: not needed anymore?
              buttonService
              .getCharacteristic(
                this.platform.Characteristic.ProgrammableSwitchOutputState,
              )
              .on(
                CharacteristicEventTypes.GET,
                this.getSwitchButtonState.bind(this, button),
              ); */
            buttonService
                .getCharacteristic(this.platform.Characteristic.ProgrammableSwitchEvent)
                .on("get" /* CharacteristicEventTypes.GET */, this.getButtonStateById.bind(this, button));
        }
        return buttonService;
    }
    getButtonStateById(button, callback) {
        const currentState = this.buttonActions[button];
        callback(this.reachabilityState, currentState);
    }
    // Button Plus can be queried (experimental API)
    // Get updated device info and update the corresponding values
    getDeviceStateUpdate() {
        return Promise.resolve();
    }
}
exports.MyStromButtonPlusAccessory = MyStromButtonPlusAccessory;
//# sourceMappingURL=myStromButtonPlusAccessory.js.map